//Librerías DS3231 (Reloj)
#include <RTClib.h>
//Librería Relés
#include <SPI.h>
//Librería DHT11 (Sensor aire)
#include <DHT.h>
//Librería Tarjeta Micro-SD
#include <SD.h>
//Librería Ethernet
#include <Ethernet.h>
//Librería Json
#include <ArduinoJson.h>
//Librería Programador
#include <TaskScheduler.h>

//definición de las funciones utilizadas
void inizialitationSensors();
void inizialitationClock();
void inizialitationScheduler();
void inizialitationEthernet();
void inizialitationSd();
void inizialitationRelay();
void loadJson(String);
void saveJson(String);
void showZones();
// void showDate();
void printDate(DateTime);
void readSensors();
void regAlarm();
void writeLog(String, String);
int *readLand(uint8_t);
void writeDate(DateTime *);
void writeDate();
// void actualDate();
void listenClient(EthernetClient *);
void readFileEthernet(String, EthernetClient *);
void postRequest(String *);
void activateZone(uint8_t *);
void deactivateZone(uint8_t *);
void readFile(String);
void removeFile(String);
void requestDateJson(String);
void requestZonesJson(String);
char *getJSONDate();
char *getJSONRead();
String getStringDate();
void checkProgram();
void software_Reset();

#define INCREMENT 2
// Configuración Relés
#define R1PIN 31 // pin del relé de la Zona 1
#define R2PIN 32 // pin del relé de la Zona 2
// Configuración Sensores
#define DHTPIN 9      // pin del sensor de humedad y temperatura del aire
#define Z1PIN A1      // pin del sensor de humedad de la Zona 1
#define Z2PIN A2      // pin del sensor de humedad de la Zona 2
#define DHTTYPE DHT11 // DHT 11
DHT dht(DHTPIN, DHTTYPE);
// Configuración DS3231 (Reloj)
RTC_DS3231 rtc;
// Configuración Micro-SD
File fichero;
// Configuración Ethernet
byte mac[] = {0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED};
IPAddress ip(192, 168, 1, 55);
EthernetServer server(80);
//Configurador Programador
Scheduler runner;
Task taskSensors(600000, TASK_FOREVER, &regAlarm);  //cada 10 minutos
Task taskCheck(60000, TASK_FOREVER, &checkProgram); //cada 1 minuto

//Estructura de datos
struct read_t //Estructura para la lectura de sensores
{
    int humZ1;
    int humZ2;
    float hum;
    float temp;
} read;
struct date //Estructura para almacenar la fecha
{
    int hour;
    int minute;
    int day;
    int month;
    int year;
} date;
struct program_t //Estructura de programación horaria
{
    int hour;
    int minute;
};
struct sensor_t //Estructura para los sensores
{
    int max;
    int min;
};
struct zone_t //Estructura para la zona
{
    // char *name;
    String name;
    program_t program[4];
    int duration;
    struct repetition_t
    {
        bool on;
        int frequency;
    } repetition;
    sensor_t air;
    sensor_t land;
    sensor_t temp;
    bool on;
};
zone_t zones[2];
struct alarm_t
{
    bool cancel = false;
    int increment = 1;
};
alarm_t alarms[2];
struct test_t
{
    bool on = false;
};
test_t test[2];

//definición de ficheros
#define WEBPAGE "index.htm"
#define CONFIG "config.txt"
#define LOG "log.txt"

void setup()
{
    Serial.begin(115200);

    //Inicialización de los diferentes módulos
    inizialitationClock();
    inizialitationRelay();
    inizialitationSensors();
    inizialitationSd();
    inizialitationEthernet();
    printDate(readDate());
    loadJson(CONFIG);
    showZones();
    delay(3000);
    inizialitationScheduler(); 
}
void loop()
{
    runner.execute();
    EthernetClient client = server.available(); // Intentando encontrar cliente web
    if (client)
    {
        listenClient(&client);
    }
}

void inizialitationSensors() //Función inicializadora de los sensores
{
    Serial.print("Inicializando sensores...");
    dht.begin();
    Serial.println(" inicialización realizada.");
}

void inizialitationClock() //Función inicializadora del RTC
{
    Serial.print("Inicializando RTC...");
    if (!rtc.begin())
    {
        Serial.print(" No se pudo encontrar el RTC");
        while (1)
            ;
    }
    if (rtc.lostPower())
    {
        writeDate(&DateTime(2018, 3, 17, 20, 0, 0));
        Serial.print(" fijando hora (pérdida de corriente)...");
    }
    Serial.println(" inicialización realizada.");
}
void inizialitationScheduler() //Función inicializadora del Programador
{
    runner.init();
    runner.addTask(taskCheck);
    runner.addTask(taskSensors);
    taskCheck.enable();
    taskSensors.enable();
}

void inizialitationEthernet() //Función inicializadora de Ethernet
{
    Serial.print("Inicializando Ethernet...");
    Ethernet.begin(mac, ip);
    server.begin();
    Serial.println(" inicialización realizada.");
}

void inizialitationSd() //Función inicializadora de la tarjeta SD
{
    Serial.print("Inicializando SD...");

    if (!SD.begin(4))
    {
        Serial.println(" inicializacion fallo!");
        return;
    }
    Serial.println(" inicialización realizada.");
}

void inizialitationRelay() //Función inicializadora de los relés
{
    Serial.print("Inicializando relés...");
    pinMode(R1PIN, OUTPUT);
    deactivateZone(R1PIN);
    pinMode(R2PIN, OUTPUT);
    deactivateZone(R2PIN);
    Serial.println(" inicialización completa.");
}

void loadJson(String filename) //JSON: Función para leer la configuración
{
    fichero = SD.open(filename);
    DynamicJsonBuffer jsonBuffer;

    JsonObject &root = jsonBuffer.parseObject(fichero);
    JsonArray &filezones = root["zones"];

    int i = 0;
    for (auto &zone : filezones)
    {
        char *aux = zone["name"];
        String aux2(aux);
        zones[i].name = aux2;
        JsonArray &programs = zone["program"];
        int j = 0;
        for (auto &program : programs)
        {
            zones[i].program[j].hour = program["hour"];
            zones[i].program[j].minute = program["minute"];
            j++;
        }
        zones[i].duration = zone["duration"];
        JsonObject &repetition = zone["repetition"];
        zones[i].repetition.on = repetition["on"];
        zones[i].repetition.frequency = repetition["frequency"];
        JsonObject &air = zone["air"];
        zones[i].air.max = air["max"];
        zones[i].air.min = air["min"];
        JsonObject &land = zone["land"];
        zones[i].land.max = land["max"];
        zones[i].land.min = land["min"];
        JsonObject &temp = zone["temp"];
        zones[i].temp.max = temp["max"];
        zones[i].temp.min = temp["min"];
        zones[i].on = zone["on"];
        i++;
    }

    fichero.close();
    jsonBuffer.clear();
}

void saveJson(String filename) //JSON: Función para guardar la configuración
{
    SD.remove(filename);
    fichero = SD.open(filename, FILE_WRITE);

    if (!fichero)
    {
        Serial.println("ERROR al crear el fichero");
        return;
    }

    DynamicJsonBuffer jsonBuffer;
    JsonObject &root = jsonBuffer.createObject();
    JsonArray &filezones = root.createNestedArray("zones");

    for (int i = 0; i < 2; i++)
    {
        JsonObject &zone = filezones.createNestedObject();
        zone["name"] = zones[i].name;
        JsonArray &program = zone.createNestedArray("program");
        int j = 0;
        while (j < 4)
        {
            JsonObject &programArray = program.createNestedObject();
            programArray["hour"] = zones[i].program[j].hour;
            programArray["minute"] = zones[i].program[j].minute;
            j++;
        }
        zone["duration"] = zones[i].duration;

        JsonObject &repetition = zone.createNestedObject("repetition");
        repetition["on"] = zones[i].repetition.on;
        repetition["frequency"] = zones[i].repetition.frequency;

        JsonObject &air = zone.createNestedObject("air");
        air["max"] = zones[i].air.max;
        air["min"] = zones[i].air.min;
        JsonObject &land = zone.createNestedObject("land");
        land["max"] = zones[i].land.max;
        land["min"] = zones[i].land.min;
        JsonObject &temp = zone.createNestedObject("temp");
        temp["max"] = zones[i].temp.max;
        temp["min"] = zones[i].temp.min;
        zone["on"] = zones[i].on;
    }
    if (root.printTo(fichero) == 0)
    {
        Serial.println("ERROR al escribir el fichero");
    }
    fichero.close();
    jsonBuffer.clear();
}
void printFile(const char *filename) //SD: Función para imprimir por serie cualquier fichero
{
    fichero = SD.open(filename);
    if (!fichero)
    {
        Serial.println(F("Failed to read file"));
        return;
    }
    while (fichero.available())
    {
        Serial.print((char)fichero.read());
    }
    Serial.println();
    fichero.close();
}

void showZones() //Función para testear el contenido de los datos
{
    Serial.println("** SHOW ZONES **");
    for (zone_t zone : zones)
    {
        Serial.print("Name: ");
        Serial.println(zone.name);
        Serial.println("Program:");
        for (program_t program : zone.program)
        {
            Serial.print("\tHour:");
            Serial.print(program.hour);
            Serial.print(" Minute:");
            Serial.println(program.minute);
        }
        Serial.print("Duration: ");
        Serial.println(zone.duration);
        Serial.print("Repetition: ");
        Serial.print("On: ");
        if (zone.repetition.on)
            Serial.print("true");
        else
            Serial.print("false");
        Serial.print(" Frequency: ");
        Serial.println(zone.repetition.frequency);
        Serial.print("Air: ");
        Serial.print("Max: ");
        Serial.print(zone.air.max);
        Serial.print(" Min: ");
        Serial.println(zone.air.min);
        Serial.print("Land: ");
        Serial.print("Max: ");
        Serial.print(zone.land.max);
        Serial.print(" Min: ");
        Serial.println(zone.land.min);
        Serial.print("Temp: ");
        Serial.print("Max: ");
        Serial.print(zone.temp.max);
        Serial.print(" Min: ");
        Serial.println(zone.temp.min);
        Serial.print("On: ");
        if (zone.on)
            Serial.println("true");
        else
            Serial.println("false");
        Serial.println();
    }
}

void showDate() //Función para testear el contenid de los datos de fecha
{
    Serial.println("** SHOW DATE **");

    Serial.print("Hora: ");
    Serial.print(date.hour);
    Serial.print(" Minutos: ");
    Serial.print(date.minute);
    Serial.print("\t Día: ");
    Serial.print(date.day);
    Serial.print(" Mes: ");
    Serial.print(date.month);
    Serial.print(" Año: ");
    Serial.println(date.year);
}

void printDate(DateTime dateNow) //RTC: Muestra la fecha almacenada en el reloj
{
    Serial.println("** PRINT DATE **");
    if (dateNow.hour() < 10)
        Serial.print('0');
    Serial.print(dateNow.hour(), DEC);
    Serial.print(':');
    if (dateNow.minute() < 10)
        Serial.print('0');
    Serial.print(dateNow.minute(), DEC);
    Serial.print(':');
    Serial.print(dateNow.second(), DEC);
    Serial.print('\t');
    if (dateNow.day() < 10)
        Serial.print('0');
    Serial.print(dateNow.day(), DEC);
    Serial.print('/');
    if (dateNow.month() < 10)
        Serial.print('0');
    Serial.print(dateNow.month(), DEC);
    Serial.print('/');
    Serial.println(dateNow.year(), DEC);
}

char *getJSONDate() //RTC: Convierte a cadena la fecha almacenada en el reloj
{
    DateTime dateNow = readDate();
    String date = "{\"hour\":";
    date += dateNow.hour();
    date += ",\"minute\":";
    date += dateNow.minute();
    date += ",\"day\":";
    date += dateNow.day();
    date += ",\"month\":";
    date += dateNow.month();
    date += ",\"year\":";
    date += dateNow.year();
    date += "}";

    char aux[70];
    date.toCharArray(aux, 70);
    return aux;
}

String getStringDate() //RTC: Convierte a cadena la fecha almacenada en el reloj
{
    DateTime dateNow = readDate();
    String date = "";
    date += dateNow.year();
    date += "/";
    if (dateNow.month() < 10)
        date += "0";
    date += dateNow.month();
    date += "/";
    if (dateNow.day() < 10)
        date += "0";
    date += dateNow.day();
    date += " ";
    if (dateNow.hour() < 10)
        date += "0";
    date += dateNow.hour();
    date += ":";
    if (dateNow.minute() < 10)
        date += "0";
    date += dateNow.minute();

    return date;
}

char *getJSONRead() //SENSORS: Convierte a cadena la fecha almacenada en el reloj
{
    readSensors();
    String r = "{\"temp\":\"";
    r += read.temp;
    r += "ºC\",\"hum\":\"";
    r += read.hum;
    r += "%\"}";

    char aux[40];
    r.toCharArray(aux, 40);
    return aux;
}

void readSensors() //SENSORS: Actualiza la lectura actual
{
    read.hum = dht.readHumidity();
    read.temp = dht.readTemperature();
    read.humZ1 = readLand(Z1PIN);
    read.humZ2 = readLand(Z2PIN);

    Serial.println("** READ SENSORS **");
}
void regAlarm()
{
    readSensors();
    for (int i = 0; i < 2; i++)
    {
        if (zones[i].on)
        {
            if (read.temp >= zones[i].temp.max)
            {
                String alarm = "<br>";
                alarm += getStringDate();
                alarm += ", TEMPERATURA: ";
                alarm += read.temp;
                alarm += ", ZONA: ";
                alarm += zones[i].name;
                alarm += ", Se aumenta el tiempo para el siguiente riego.";
                alarm += "</br>";
                alarms[i].increment = INCREMENT;
                writeLog(LOG, alarm);
            }
            if (read.temp <= zones[i].temp.min)
            {
                String alarm = "<br>";
                alarm += getStringDate();
                alarm += ", TEMPERATURA: ";
                alarm += read.temp;
                alarm += ", ZONA: ";
                alarm += zones[i].name;
                alarm += ", Se cancela siguiente riego.";
                alarm += "</br>";
                alarms[i].cancel = true;
                writeLog(LOG, alarm);
            }
            if (read.hum >= zones[i].air.max)
            {
                String alarm = "<br>";
                alarm += getStringDate();
                alarm += ", HUMEDAD: ";
                alarm += read.hum;
                alarm += ", ZONA: ";
                alarm += zones[i].name;
                alarm += ", Se cancela siguiente riego.";
                alarm += "</br>";
                alarms[i].cancel = true;
                writeLog(LOG, alarm);
            }
            if (read.hum <= zones[i].air.min)
            {
                String alarm = "<br>";
                alarm += getStringDate();
                alarm += ", HUMEDAD: ";
                alarm += read.hum;
                alarm += ", ZONA: ";
                alarm += zones[i].name;
                alarm += ", Se aumenta el tiempo para el siguiente riego.";
                alarm += "</br>";
                alarms[i].increment = INCREMENT;
                writeLog(LOG, alarm);
            }
        }
    }
    if (zones[0].on)
    {
        if (read.humZ1 >= zones[0].land.max)
        {
            String alarm = "<br>";
            alarm += getStringDate();
            alarm += ", HUMEDAD TIERRA: ";
            alarm += read.humZ1;
            alarm += ", ZONA: ";
            alarm += zones[0].name;
            alarm += ", Se aumenta el tiempo para el siguiente riego.";
            alarm += "</br>";
            alarms[0].increment = INCREMENT;
            writeLog(LOG, alarm);
        }
        if (read.humZ1 <= zones[0].land.min)
        {
            String alarm = "<br>";
            alarm += getStringDate();
            alarm += ", HUMEDAD TIERRA: ";
            alarm += read.humZ1;
            alarm += ", ZONA: ";
            alarm += zones[0].name;
            alarm += ", Se cancela siguiente riego.";
            alarm += "</br>";
            alarms[0].cancel = true;
            writeLog(LOG, alarm);
        }
    }
    if (zones[1].on)
    {
        if (read.humZ2 >= zones[1].land.max)
        {
            String alarm = "<br>";
            alarm += getStringDate();
            alarm += ", HUMEDAD TIERRA: ";
            alarm += read.humZ2;
            alarm += ", ZONA: ";
            alarm += zones[1].name;
            alarm += ", Se aumenta el tiempo para el siguiente riego.";
            alarm += "</br>";
            alarms[1].increment = INCREMENT;
            writeLog(LOG, alarm);
        }
        if (read.humZ2 <= zones[1].land.min)
        {
            String alarm = "<br>";

            alarm += getStringDate();
            alarm += ", HUMEDAD TIERRA: ";
            alarm += read.humZ2;
            alarm += ", ZONA: ";
            alarm += zones[1].name;
            alarm += ", Se cancela siguiente riego.";
            alarm += "</br>";
            alarms[1].cancel = true;
            writeLog(LOG, alarm);
        }
    }
}

void writeLog(String filename, String alarm) //SD: Escritura de fichero de registro
{
    fichero = SD.open(filename, FILE_WRITE);
    if (fichero)
    {
        fichero.println(alarm);
        fichero.close();
    }
    else
    {
        Serial.println("ERROR abriendo " + filename);
    }
}

int *readLand(uint8_t pin) //SENSORS: Lectura del sensor de humedad de tierra
{
    return analogRead(pin);
}

void writeDate(DateTime *newDate) //RTC: Almacena la fecha especificada en el reloj
{
    rtc.adjust(*newDate);
}
void writeDate() //RTC: Almacena la fecha de la estructura en el reloj
{
    rtc.adjust(DateTime(date.year, date.month, date.day, date.hour, date.minute, 0));
}

DateTime readDate() //RTC: Devuelve la fecha almacenada en el reloj
{
    return rtc.now();
}

void actualDate() //RTC: Almacena la fecha actual en la estructura de datos
{
    DateTime now = readDate();
    date.hour = now.hour();
    date.minute = now.minute();
    date.day = now.day();
    date.month = now.month();
    date.year = now.year();
}

void listenClient(EthernetClient *client) //ETHERNET: Escuha las peticiones de clientes
{
    String req_str = "";
    boolean currentLineIsBlank = true;

    while (client->connected())
    {
        if (client->available())
        {
            char c = client->read();
            req_str += c;
            // Serial.print(c_s);

            if (c == '\n' && currentLineIsBlank)
            {
                // Se envía cabecera de respuesta estándar
                client->println("HTTP/1.1 200 OK");
                client->println("Content-Type: text/html");
                client->println("Connection: close");
                client->println();
                // Envío de página web
                // Serial.print(req_str);
                if (req_str.indexOf("POST") >= 0)
                {
                    postRequest(&req_str);
                    break;
                }
                else
                {
                    if (req_str.indexOf("GET / ") >= 0)
                    {
                        readFileEthernet(WEBPAGE, client);
                    }
                    if (req_str.indexOf("GET /read") >= 0)
                    {
                        client->write(getJSONRead());
                    }
                    if (req_str.indexOf("GET /date") >= 0)
                    {
                        client->write(getJSONDate());
                    }
                    if (req_str.indexOf("GET /zones") >= 0)
                    {
                        readFileEthernet(CONFIG, client);
                    }
                    if (req_str.indexOf("GET /log ") >= 0)
                    {
                        readFileEthernet(LOG, client);
                    }
                    if (req_str.indexOf("GET /removeLog ") >= 0)
                    {
                        removeFile(LOG);
                    }
                    if (req_str.indexOf("PUT /zone1 ") >= 0)
                    {
                        if (test[0].on)
                        {
                            test[0].on = false;
                            deactivateZone(R1PIN);
                        }
                        else
                        {
                            test[0].on = true;
                            activateZone(R1PIN);
                        }
                    }
                    if (req_str.indexOf("PUT /zone2 ") >= 0)
                    {
                        if (test[1].on)
                        {
                            test[1].on = false;
                            deactivateZone(R2PIN);
                            return;
                        }
                        else
                        {
                            test[1].on = true;
                            activateZone(R2PIN);
                        }
                    }

                    break;
                }
            }
            // Cada línea de texto recibida del cliente termina con \r\n
            if (c == '\n')
            {
                // último caracter de la línea del texto recibido
                // empezando nueva linea con el siguiente caracter leído
                currentLineIsBlank = true;
            }
            else if (c != '\r')
            {
                // un caracter de texto fue recibido desde el cliente
                currentLineIsBlank = false;
            }
        }
    }
    delay(1);       // da al navegador web tiempo para recibir los datos
    client->stop(); // cierre de la conexión
}

void readFileEthernet(String filename, EthernetClient *client) //SD: Devuelve al cliente de ethernet el fichero de datos
{
    fichero = SD.open(filename);
    if (fichero)
    {
        while (fichero.available())
        {
            client->write(fichero.read());
        }
        fichero.close();
    }
    else
    {
        char *error = "ERROR abriendo fichero ";
        Serial.print(error);
        Serial.println(filename);
        client->write(error);
    }
}

void postRequest(String *req_str) //ETHERNET: Limpia los datos recogidos desde el cliente
{
    String data = req_str->substring(6, req_str->indexOf("HTTP"));
    data.replace("%7B", "{");
    data.replace("%22", "\"");
    data.replace("%7D", "}");
    data.replace("%20", " ");
    if (data.indexOf("year") >= 0)
    {
        requestDateJson(data);
    }
    else
    {
        requestZonesJson(data);
    }
}

void activateZone(uint8_t *zone) //RELAY: Activa la zona indicada
{
    digitalWrite(zone, LOW);
}
void deactivateZone(uint8_t *zone) //RELAY: Desactiva la zona indicada
{
    digitalWrite(zone, HIGH);
}

void readFile(String filename) //SD: Lectura de fichero
{
    fichero = SD.open(filename);
    if (fichero)
    {
        Serial.println("Contenido del fichero " + filename + ":");

        while (fichero.available())
        {
            Serial.write(fichero.read());
        }
        fichero.close();
    }
    else
    {
        Serial.println("ERROR abriendo " + filename);
    }
}

void requestDateJson(String data) //JSON: Almacenar datos en la estructura
{
    Serial.println("** REQUEST DATE **");

    DynamicJsonBuffer jsonBuffer;
    JsonObject &root = jsonBuffer.parseObject(data);
    date.year = root["year"];
    date.month = root["month"];
    date.day = root["day"];
    date.hour = root["hour"];
    date.minute = root["minute"];

    writeDate();           //Alnacena en RTC
    printDate(readDate()); //Muestra RTC

    jsonBuffer.clear();
}
void requestZonesJson(String data)
{
    Serial.println("** REQUEST ZONES **");

    DynamicJsonBuffer jsonBuffer;
    JsonObject &root = jsonBuffer.parseObject(data);

    int i = 0;
    String a = "";
    switch (data[2])
    {
    case '1':
        i = 0;
        a = "1";
        break;
    case '2':
        i = 1;
        a = "2";
        break;
    }

    JsonObject &zone = root[a];
    char *aux = zone["name"];
    String aux2(aux);
    zones[i].name = aux2;
    JsonArray &programs = zone["program"];
    int j = 0;
    for (auto &program : programs)
    {
        zones[i].program[j].hour = program["hour"];
        zones[i].program[j].minute = program["minute"];
        j++;
    }
    zones[i].duration = zone["duration"];
    JsonObject &repetition = zone["repetition"];
    zones[i].repetition.on = repetition["on"];
    zones[i].repetition.frequency = repetition["frequency"];
    JsonObject &air = zone["air"];
    zones[i].air.max = air["max"];
    zones[i].air.min = air["min"];
    JsonObject &land = zone["land"];
    zones[i].land.max = land["max"];
    zones[i].land.min = land["min"];
    JsonObject &temp = zone["temp"];
    zones[i].temp.max = temp["max"];
    zones[i].temp.min = temp["min"];
    zones[i].on = zone["on"];

    jsonBuffer.clear();

    saveJson(CONFIG);

    if (i == 1)
        software_Reset();
}
void removeFile(String filename)
{
    Serial.println("** REMOVE LOG **");
    SD.remove(filename);
}

void checkProgram()
{
    DateTime now = readDate();

    for (int i = 0; i < 2; i++)
    {
        if (zones[i].on)
        {
            if (!zones[i].repetition.on)
            {
                for (program_t program : zones[i].program)
                {
                    if (program.hour == now.hour() && program.minute == now.minute())
                    {
                        if (alarms[i].cancel)
                        {
                            alarms[i].cancel = false;
                            return;
                        }

                        Serial.print("** Encenciendo ");
                        Serial.print(zones[i].name);
                        Serial.println(" **");

                        if (i == 0)
                            activateZone(R1PIN);
                        if (i == 1)
                            activateZone(R2PIN);
                    }
                    if (program.hour == now.hour() && ((zones[i].duration * alarms[i].increment) + program.minute) == now.minute())
                    {
                        Serial.print("** Apagando ");
                        Serial.print(zones[i].name);
                        Serial.println(" **");

                        if (i == 0)
                            deactivateZone(R1PIN);
                        if (i == 1)
                            deactivateZone(R2PIN);
                        alarms[i].increment = 1;
                    }
                }
            }
            else
            {
                if ((now.minute() - zones[i].program[0].minute) % zones[i].repetition.frequency == 0)
                {
                    if (alarms[i].cancel)
                    {
                        alarms[i].cancel = false;
                        return;
                    }

                    Serial.print("** Encenciendo ");
                    Serial.print(zones[i].name);
                    Serial.println(" **");

                    if (i == 0)
                        activateZone(R1PIN);
                    if (i == 1)
                        activateZone(R2PIN);
                }
                if ((now.minute() - zones[i].program[0].minute - (zones[i].duration * alarms[i].increment)) % zones[i].repetition.frequency == 0)
                {
                    Serial.print("** Apagando ");
                    Serial.print(zones[i].name);
                    Serial.println(" **");

                    if (i == 0)
                        deactivateZone(R1PIN);
                    if (i == 1)
                        deactivateZone(R2PIN);
                    alarms[i].increment = 1;
                }
            }
        }
    }
}

void software_Reset()
{
    Serial.println("** REINICIANDO **");
    delay(3000);
    asm volatile(" jmp 0");
}